#!/usr/bin/perl
# vim: set syntax=perl tabstop=4 foldmethod=marker:

# asyd-ca.pl:
# 		Perl script to manage a little CA using CLI
#
# Bruno Bonfils, <asyd@asyd.net>
# 2006, January
# Published under term of BSD licence

use strict;
use warnings qw(all);

my $command = shift @ARGV;

# configuration. You can tweak it
# {{{ hash config
my $config = {
	# Where is the 
	openssl => '/usr/bin/openssl',
	# Key infos
	key => {
		# Valid type are rsa and dsa
		type => 'rsa',
		size => 1024,
	},
	# Debug
	debug => 0,
	# Verbose
	verbose => 1,
	firstserial => '000000',
};
# }}}

# {{{ main
help('root', 1) unless (defined $command);

{
	my $buffer = $0;
	$buffer =~ s#(.*/).+#$1#;
	$config->{'basedir'}  = $buffer;
}

if ($command eq 'help') {
	if (defined $ARGV[0]) {
		help ($ARGV[0], 0);	
	} else {
		help ('root', 0);	
	}
} elsif ($command eq 'init') {
	init(@ARGV);
} elsif ($command eq 'purge') {
	purge(@ARGV);
} elsif ($command eq 'req') {
	req(@ARGV);
} elsif ($command eq 'delete') {
	attic(@ARGV);
} elsif ($command eq 'sign') {
	sign(@ARGV);
} elsif ($command eq 'pkcs12') {
	pkcs12(@ARGV);
} else {
	help ('root', 1);
}
# }}}

# sub function
# {{{ sub pkcs12
sub pkcs12 {
	my ($name) = @_;
	help ('pkcs12', 1) unless (defined $name);
	my @cmd = ( 'pkcs12', '-export',
                '-CAfile', $config->{'basedir'} . "CA/cacert.pem",
                '-out', $config->{'basedir'} . "pkcs12/$name.p12",
                '-in', $config->{'basedir'} . "certs/$name.pem",
                '-inkey', $config->{'basedir'} . "keys/$name.key",
                '-caname', 'root', '-chain',
                '-name', $name);
	run_cmd (@cmd);
}
# }}}
# {{{ sub sign
sub sign {
	my ($name) = @_;
	help ('sign', 1) unless (defined $name);
	my @cmd;
	@cmd = ( 'ca',
             '-policy', 'policy_anything',
             '-config', $config->{'basedir'} . "templates/ca.cnf",
             '-in', $config->{'basedir'} . "reqs/$name.req",
             '-out', $config->{'basedir'} . "certs/$name.pem" );
	run_cmd (@cmd);	
}
# }}}
# {{{ sub attic (Move files to attic)
sub attic {
	my ($name) = @_;
	help ('delete', 1) unless (defined $name);

	# Timestamp
	my ($day, $month, $year, $hour, $minute) = (localtime)[3,4,5,2,1];
	my $timestamp = (1900 + $year) . sprintf("%02d", $month + 1)  . sprintf("%02d", $day) . '-' . $hour . ':' . $minute;

	rename $config->{'basedir'} . "keys/$name.key", $config->{'basedir'} . '/misc/attic/' . "$name-$timestamp.key";
	rename $config->{'basedir'} . "reqs/$name.req", $config->{'basedir'} . '/misc/attic/' . "$name-$timestamp.req";
	rename $config->{'basedir'} . "certs/$name.pem", $config->{'basedir'} . '/misc/attic/' . "$name-$timestamp.pem";
	rename $config->{'basedir'} . "pkcs12/$name.p12", $config->{'basedir'} . '/misc/attic/' . "$name-$timestamp.p12";
}
# }}}
# {{{ sub req
sub req {
	my ($name, $template) = @_; 
	unless (defined $name) {
		help('req', 1);
	}

	$template = 'simple.cnf' unless (defined $template);

	my @cmd;
	unless (-r "keys/$name.key") {
		push @cmd, 'gen' . $config->{'key'}->{'type'}, '-out', "keys/$name.key", $config->{'key'}->{'size'};
		run_cmd (@cmd);
	}

	@cmd = ();
	push @cmd, 'req', '-config', 'templates/' . $template, '-new', '-out', "reqs/$name.req", '-key', "keys/$name.key";
	run_cmd (@cmd);
}
# }}}
# {{{ sub purge 
sub purge {
	my @cmd = ('rm', '-fr', 
               $config->{'basedir'} . 'CA',
               $config->{'basedir'} . 'certs',
               $config->{'basedir'} . 'keys', 
               $config->{'basedir'} . 'misc', 
               $config->{'basedir'} . 'pkcs12', 
               $config->{'basedir'} . 'reqs' );
	qx!@cmd!;
}
# }}}
# {{{ sub run_cmd
sub run_cmd {
	my @args;
	push @args, $config->{'openssl'};
	push @args, @_;

	# Print which command is execute if verbose mode is enable

	#system(@args) or sub { warn "system @args failed: $?" if ($config->{'debug'}); };
	my $buffer = qx!@args!;

	my $ret = $? >> 8;
	if ($config->{'verbose'}) {
		print "Return value: $ret, command: ", join(' ', @args), "\n";
	}
	return $ret;
} # }}}
# {{{ sub init
sub init {
	my @cmd;
	mkdir $config->{'basedir'} . 'CA' , 0700;
	mkdir $config->{'basedir'} . 'misc' , 0700;
	mkdir $config->{'basedir'} . 'misc/attic' , 0700;
	mkdir $config->{'basedir'} . 'keys', 0700;
	mkdir $config->{'basedir'} . 'certs' , 0700;
	mkdir $config->{'basedir'} . 'reqs' , 0700;
	mkdir $config->{'basedir'} . 'pkcs12' , 0700;
	my $fh; open ($fh, "> " . $config->{'basedir'} . "misc/serial"); print $fh $config->{'firstserial'} . "\n"; close ($fh);
	open ($fh, "> " . $config->{'basedir'} . 'misc/index.txt'); close ($fh);

	# Compute CA key
	@cmd = ();
	push @cmd, 'gen' . $config->{'key'}->{'type'}, '-out', $config->{'basedir'} . 'CA/ca.key', $config->{'key'}->{'size'};
	run_cmd (@cmd);
	# Request
	@cmd = ('req', 
            '-new',
            '-config', $config->{'basedir'} . 'templates/ca-req.cnf',
            '-out', $config->{'basedir'} . 'CA/cacert.req',
            '-key', $config->{'basedir'} . 'CA/ca.key');
	run_cmd (@cmd);
	# Self sign and init some openssl stuff
	@cmd = ();
	push @cmd, 'ca',
               '-config', $config->{'basedir'} . 'templates/ca.cnf',
               '-out', $config->{'basedir'} . 'CA/cacert.pem',
               '-batch', 
               '-keyfile', $config->{'basedir'} . 'CA/ca.key',
               '-selfsign',
               '-infiles',
               'CA/cacert.req';
	run_cmd (@cmd);
	unlink $config->{'basedir'} . 'CA/cacert.req';
}
# }}}
# {{{ sub help
sub help {
	my ($context, $code) = @_;

	if ($context eq 'init') {
		print "init\n";
	} elsif ($context eq 'req') {
		print "usage: $0 req name [template]\n";
		print "\treq      Create a certificate request\n";
        print "\tname     Shortname (required)\n";
		print "\ttemplate Template to use (just the name file, like wildcard.cnf), optional\n";
		print "\texemple: $0 req web01-apache\n";
	} elsif ($context eq 'delete') {
		print "usage: $0 delete name\n";
		print "\tdelete   Move all files related to a cert to attic\n"; 
	} elsif ($context eq 'sign') {
		print "usage: $0 sign name\n";
		print "\tsign     Sign a request\n";
	} elsif ($context eq 'pkcs12') {
		print "usage: $0 pkcs12 name\n";
		print "\tpkcs12   Create a PKCS12 file from cacert, cert and key\n";
	} else {
		print "usage: $0 subcommand [options]\n";
		print "Sub command :\n";
		print "\thelp     Display this message\n";
		print "\tinit     Create CA certs and requirements\n";
		print "\treq      Create a certificate request\n";
		print "\tsign     Sign a request\n";
		print "\tdelete   Move all files related to a cert to attic\n"; 
		print "\tpkcs12   Create a PKCS12 file from cacert, cert and key\n";
		print "\tpurge    Purge ALL files (dangerous !)\n";
	}
	
	exit ($code) if ($code > 0);
}
# }}}
